/*
 ============================================================================
 Project     : Standard Genetic Algorithm
 Author      : Lucas Angelo Silveira
 Universidade de Brasilia (UnB) - Brazil
 Group of Theory of Computation
 Version     : 1.0
 
 Last Modified On: July 16, 2015.
 ============================================================================
 */



#include <stdio.h>
#include <stdlib.h> 
#include "structure.h"
#include <math.h> //log2
#include "../Fitness/Processing.c"
#include "Alg_Simple_Genetic.h"

/*
const float prob_crossover = 0.98;     //Crossover Probability
const float porc_selection = 0.96;     //% of the population that will be selected for Crossover
const float porc_reemp= 1 - 0.6;  //Pct. of the population from where it will be performed
const float prob_mutation= 0.01;
*/

float prob_crossover;     //Crossover Probability
float porc_selection;     //% of the population that will be selected for Crossover
float porc_reemp;  //Pct. of the population from where it will be performed
float prob_mutation;     //Mutation Probability
int * identity;
int size;
int sentinela;
int *c;
Genome *a;

void read_permutation_of_archivo(char* name, permutation *perm)
{
    int i,n;
    FILE *filePointer;
    if ((filePointer = fopen(name,"r")) == NULL){
        printf("Warning:Filen not found\n");
    }
    
    else
    {
        //read size of permutation
        fscanf(filePointer,"%d",&n);
        perm->size = n;
        size=n;
        //alocate memory for perm->pi
        perm->pi = malloc(perm->size*sizeof(int));
        //leer permutacion
        for(i=0; i<n; i++)
        {
            fscanf(filePointer,"%d",&perm->pi[i]);
        }
        fscanf(filePointer,"%d",&n);
        identity= (int *) calloc (n, sizeof (int));
        for(i=0; i<n; i++)
        {
            fscanf(filePointer,"%d",&identity[i]);
        }
        
        
    }//end  else
    fclose(filePointer);
}//end read permutation of archivo



void see_population(permutation *perm,population *pob)
{
 int i,j;
  printf("mostrando a populacao gerada em forma de permutacao\n");
 for(i=0; i<pob->sizePopulation;i++ )
  {
	for(j=0;j<perm->size;j++)
	{
            printf("%d,", pob->genome[i].genes[j]);
		
	}
     printf("\n");
  }
}//end see_population


// Procedure that generates the initial population
 
void population_init(population *pob, permutation *perm){
	int i,j,r,n,prob;
	
    //Calc size of population
	n = perm->size;
	pob->sizePopulation =  (int) (n * (log(n) / log(2))); //siz of population = (nlogn); log in base 2
	pob->basePopulation = 0;
	pob->topPopulation = pob->sizePopulation;
    
	pob->limiteSelection = pob->sizePopulation * porc_selection;//limit de population what go to crossover
	pob->baseReplace = pob->sizePopulation * porc_reemp;//base de pob a partir de where will replace
    
	pob->baseDescendencia = pob->sizePopulation;
	pob->topeDescendencia = pob->sizePopulation + pob->limiteSelection;
    
	pob->sizeTotal = pob->topeDescendencia;// tamanho total = poblacion + descendencia
    
    /*alocar memoria*/
	pob->genome = malloc(pob->sizeTotal*sizeof(Genome)); //alocar memoria para poblacion total
    if (pob->genome == NULL)
        printf("pob->genome NULL in population_init(x,x) in Alg_Geneti_simple\n");
    
    for(i=0;i<pob->sizeTotal;i++)
    {
        pob->genome[i].genes = malloc(n * sizeof(int));//alocar memoria para cada genome
        if (pob->genome[i].genes == NULL)
            printf("pob->genome[i].genes NULL in population_init(x,x)\n");
    }
    
	/*create population a partir of  permutacion*/
       // A genom is see with a permutation of concatenation chroms 
        
    /*create population a partir of  permutacion*/
    // A genome is see with a permutation of concatenation chroms
    
	for(i=pob->basePopulation; i < pob->topPopulation; i++)
    {
        
		for(j = 0; j < perm->size; j++)
        {
            if(perm->pi[j]!=0 && j!=0 && (j+1 < perm->size) && perm->pi[j+1]!=0 && perm->pi[j-1]!=0  )
            {
                r=rand();
                prob = r % 2;//valores entre 0 y 1
                if (prob == 1)//si prob 1, entonces gen positivo
                {
                    pob->genome[i].genes[j] = perm->pi[j];
                    
                }
                else//si prob 0, entonces gen negativo
                {
                    
                    pob->genome[i].genes[j] = perm->pi[j] * -1;
                    
                }
            }//end if
            else
            {
                pob->genome[i].genes[j] = perm->pi[j];
                
            }
        }//end for interno
        
		pob->genome[i].size = perm->size;
	}//end for extern
	/*agregar tamanho a desc*/
	for(i=pob->baseDescendencia; i<pob->topeDescendencia; i++){
		pob->genome[i].size = perm->size;
	}

}//end create init population


//flag(0) indica se e um populacao inicial ou (1) se ja contem descentes
void see_fitness(population *pob ,int flag)
{
    int i;
    if(!flag)
    {
        for(i=pob->basePopulation; i < pob->sizePopulation; i++)
            printf("%d: distancia eh:%d\n",i, pob->genome[i].fitness);
    }
    else
    {
        
        for(i=pob->baseDescendencia; i < pob->topeDescendencia; i++)
            printf("distancia eh:%d\n",pob->genome[i].fitness);
    }
    
}

//compute the distance to each genome of the population
void fitness_population(population *pob, int flag){
	int i;
    
	if (flag > 1)
	{ //fitness sobre descendencia
        
        
        
		for(i=pob->baseDescendencia; i < pob->topeDescendencia; i++)
		{
            

		 	load_genoma( pob->genome[i].genes, identity,pob->genome[i].size, size);
            
            
            pob->genome[i].fitness =distance(pob->genome[i].genes,identity, pob->genome[i].size,2);
            
            
        }
        
	}
	else
    { // fitness sobre poblacion
        for(i=pob->basePopulation; i < pob->sizePopulation; i++)
        {
       
            load_genoma( pob->genome[i].genes, identity,pob->genome[i].size, size);
            pob->genome[i].fitness = distance(pob->genome[i].genes,identity, pob->genome[i].size,2);
            
        }
    }
}


void countingSort(population *pob, int k){
	int i,j;
	int *c;//c[k];
	Genome *a;
    
    /*alocar memoria para c[k]*/
    c = malloc(k*sizeof(int));
    
	a = malloc(pob->sizePopulation*sizeof(Genome));
	if (a == NULL) printf("a NULL en countingSort() \n");
    
    for(i=0;i<pob->sizePopulation;i++){
        a[i].genes = malloc(pob->genome[0].size * sizeof(int));//alocar memoria para cada cromosoma
        if (a[i].genes == NULL)
            printf("a[i].genes NULL en countingSort()\n");
    }
    
	for(i=0; i<pob->sizePopulation; i++){
		a[i].fitness = pob->genome[i].fitness;
        for(j=0; j<pob->genome[0].size; j++)
            a[i].genes[j] = pob->genome[i].genes[j];
	}
    
	/*counting sort*/
	for(i=0; i<k; i++){
		c[i] = 0;
	}
    
	for(i=0; i<pob->sizePopulation; i++){
		c[a[i].fitness] = c[a[i].fitness] + 1;
	}
    
	for(i=1; i<k; i++){
		c[i] = c[i] + c[i-1];
	}
    
	for(i=pob->sizePopulation-1; i>=0; i--){
		
        pob->genome[c[a[i].fitness]-1].fitness = a[i].fitness;
        for(j=0; j<pob->genome[0].size; j++)
            pob->genome[c[a[i].fitness]-1].genes[j] = a[i].genes[j];
        
        
		c[a[i].fitness] = c[a[i].fitness] - 1;
	}
    
    
    /*Liberar memoria*/
    for(i=0;i<pob->sizePopulation;i++){
        free(a[i].genes);
    }
	free(a);
    free(c);
}


void selection(population *pob,  int n_generetion, int * best_solution){
	
    countingSort(pob,pob->genome[0].size+2);
    
    if (pob->genome[0].fitness < *best_solution)
        *best_solution = pob->genome[0].fitness;
    
}


void crossover(population *pob, permutation *perm){
	int i,j,k,pos_pai1, pos_pai2, point_crossover;
	float prob;
	k = pob->baseDescendencia;//indice para colocar en descendencia
	/*crossver sobre populacao atual*/
	for(i=pob->basePopulation; i<pob->limiteSelection; i+=2)
    {
		//se falta ultima posicao para colocar descendencia, terminar
		if (k+1 == pob->sizeTotal)
			break;
        
		//escolher posicao dos pais
		pos_pai1 = rand() % pob->limiteSelection;//valores entre 0 e limiteSelection-1
		pos_pai2 = rand() % pob->limiteSelection;//valores entre 0 e limiteSelection-1
		//buscar pos2 que no seja igual a pos1, pelo  menos 10vezes
		for(j=0; j<10; j++){
			if (pos_pai1 != pos_pai2)
                break;
			pos_pai2 = rand() % pob->limiteSelection;//valores entre 0 e limiteSelection-1
		}
        
		prob = ((rand() % 100) + 1)/100.0; //de 0.01, 0.02 ... a 1.0
        
		if (prob <= prob_crossover)
        {//crossover
			point_crossover = rand() % perm->size;//valores entre 0 e tamanho de permutation - 1
			//aplicar single_point_crossover na primera parte antes do ponto de crossover
			for(j=0; j<point_crossover; j++){
				pob->genome[k].genes[j] = pob->genome[pos_pai1].genes[j];
				pob->genome[k+1].genes[j] = pob->genome[pos_pai2].genes[j];
			}
			//aplicar single_point_crossover na segunda parte depois do ponto de crossover
			for(j=point_crossover; j<perm->size; j++){
				pob->genome[k+1].genes[j] = pob->genome[pos_pai1].genes[j];
				pob->genome[k].genes[j] = pob->genome[pos_pai2].genes[j];
			}
			k = k + 2;
		}//end if crossover
	}//end for iniit
	pob->topeDescendencia = k;//atualizar tope descendencia
}


void mutation(population *pob, permutation *perm){
	int i,j;
	float prob;
    
	/*mutation about descendents, baseDescendencia=0*/
	for(i=pob->baseDescendencia; i<pob->topeDescendencia; i++)
    {
		for(j=0; j<perm->size; j++)
        {
			prob = ((rand() % 100) + 1)/100.0; //de 0.01, 0.02 ... a 1.0
			
            if(prob <= prob_mutation)
            {//applied a gene mutation (flip sign)
                
                if(pob->genome[i].genes[j]!=0 && j!=0 && (j+1 < size) && pob->genome[i].genes[j+1]!=0 && pob->genome[i].genes[j-1]!=0 )
                    pob->genome[i].genes[j] = pob->genome[i].genes[j] * -1;
                
            }
            
        }
        
    }//end for
    
}//end of the function mutation



void replacement(population *pob){
	int i,j,pos_reemp;
    int conta=0;
	/*substituir descendentes na populacao atual*/
    for(i=pob->baseDescendencia; i<pob->topeDescendencia; i++)
    {
		//buscar pos de substituicao
		int tamanhoReemp = pob->topPopulation - pob->baseReplace;
		pos_reemp = (rand() % tamanhoReemp) + pob->baseReplace;//(valores entre 0 e tamanhoReemp-1)+baseReplace
		
        //realizar substituicao
		if (pob->genome[i].fitness < pob->genome[pos_reemp].fitness)
        {
            conta++;
			//copiar  fitness
            pob->genome[pos_reemp].fitness = pob->genome[i].fitness;
            //copiar genes
            for(j=0; j<pob->genome[0].size; j++)
                pob->genome[pos_reemp].genes[j] = pob->genome[i].genes[j];
            
        }
	}
}



void SimpleGenetic(float probcrossover, float probmutacao, float probselecao, float probsubstituicao)
{
    
 int i=0;
 permutation perm;
 sentinela=0;
 population pob;
 int best_solution;
 int num_generation;
 read_permutation_of_archivo("GENOME", &perm);// leitura validada
 //verifica se is-cotail
    if (is_cotail(perm.pi,identity,perm.size, size))
    {
        prob_crossover=0.98;
        prob_mutation=0.018;
        porc_selection=0.8;
        porc_reemp=1.0 - 0.7;
        
        //see_permutation(&perm);
        population_init(&pob, &perm);
        
        set_TAMANHO_CAPPEPLUS( perm.pi, perm.size);
        //function of Preprocessamento.c
        load_nodeToCompMap();

       // see_population(&perm,&pob);

        fitness_population(&pob, 1);
       // see_fitness(&pob, 0);
        best_solution = pob.genome[0].fitness;
        num_generation = perm.size;
        
        for(i=2; i<=num_generation; i++)
        {
            
            selection(&pob,i,&best_solution);//seleccion:ordena populacao
            crossover(&pob, &perm);
           // see_population(&perm,&pob);
            mutation(&pob, &perm);
            fitness_population(&pob,i);
            //see_fitness(&pob, 0);
            replacement(&pob);
        }
     
    
        selection(&pob,i,&best_solution);//seleccion:ordena populacao
       // printf("\n");
        //see_population(&perm,&pob);
        printf("Distance Standard GA:%d\n", best_solution);
        
    }
    else
        printf("*****************INPUT NOT IS-COTAIL************************");
    
    
}//end function Genetico




